#!/bin/bash
#

# build_scn.sh <product_dir> <quals> <debug|prof> [tar]

usage()
{
   echo "USAGE: `basename ${0}` <product_dir> <quals> <debug|prof> [tar]"
}

# -------------------------------------------------------------------
# shared boilerplate
# -------------------------------------------------------------------

get_this_dir() 
{
    ( cd / ; /bin/pwd -P ) >/dev/null 2>&1
    if (( $? == 0 )); then
      pwd_P_arg="-P"
    fi
    reldir=`dirname ${0}`
    thisdir=`cd ${reldir} && /bin/pwd ${pwd_P_arg}`
}

get_ssibuildshims()
{
    # make sure we can use the setup alias
    if [ -z ${UPS_DIR} ]
    then
       echo "ERROR: please setup ups"
       exit 1
    fi
    source `${UPS_DIR}/bin/ups setup ${SETUP_UPS}`

    cd ${product_dir}
    echo "setup ssibuildshims ${ssibuildshims_version} -z ${product_dir}"
    setup ssibuildshims ${ssibuildshims_version} -z ${product_dir}
}

# -------------------------------------------------------------------
# start processing
# -------------------------------------------------------------------

product_dir=${1}
#ignore_basequal=${2}
#ignore_extraqual=${3}
maketar=${2}

if [ -z ${product_dir} ]
then
   echo "ERROR: please specify the local product directory"
   usage
   exit 1
fi

# -------------------------------------------------------------------
# package name and version
# -------------------------------------------------------------------

package=scn              # Ups package name
origpkgver=v01_00_00     # Git tag
pkgver=v01_00_00         # Ups version
ssibuildshims_version=v2_17_01
pkgtarname=${package}-${origpkgver}    # Source code tarball name

# Pip versions.
# To view available versions, visit https://pypi.org/project/<pip-package>

torch_version=2.6.0
torchvision_version=0.21.0

get_this_dir

get_ssibuildshims

source ssi_functions
define_basics NO_QUAL --

#if [ "${maketar}" = "tar" ]
#then
#   ${SSIBUILDSHIMS_DIR}/bin/make_distribution_tarball ${product_dir} ${package} ${pkgver} ${fullqual}
#   exit 0
#fi

echo "package = $package"
echo "pkgver = $pkgver"
echo "product_dir = $product_dir"
echo "fq_dir = $fq_dir"
echo "fullqual = $fullqual"
echo "OS = $OS"
echo "plat = $plat"
echo "flvr = $flvr"
echo "pkgdir = $pkgdir"
echo "pkgtop = $pkgtop"
echo "maketar = $maketar"
#echo "building ${package} for ${OS}-${plat}-${qualdir} (flavor ${flvr})"

# Make build script reentrant

rm -rf $pkgdir
rm -rf $pkgtop/fakedb
rm -rf $product_dir/$package/${pkgver}.version

# -------------------------------------------------------------------
# package dependent stuff goes here
# -------------------------------------------------------------------

#ncore=`${SSIBUILDSHIMS_DIR}/bin/ncores`

mkdir -p ${pkgdir}
if [ ! -d ${pkgdir} ]
then
   echo "ERROR: failed to create ${pkgdir}"
   exit 1
fi

# fake ups declare
fakedb=${product_dir}/${package}/${pkgver}/fakedb
${SSIBUILDSHIMS_DIR}/bin/fake_declare_product ${product_dir} ${package} ${pkgver} NO_QUAL

# setup fake package and dependencies (python).
cd ${product_dir}
setup -B ${package} ${pkgver} -z ${fakedb}:${product_dir}:${PRODUCTS} || die "ERROR: fake setup failed"

# setup gcc

setup gcc v12_1_0

set -x
cd ${pkgdir} || exit 1

# Make python virtual virtual environment.

python -m venv venv
source venv/bin/activate
pip install torch==${torch_version} torchvision==${torchvision_version}

# Update activate script and other scripts in virtual environment.

find venv/bin -type f -exec grep -l $pkgdir {} \; | while read f
do
  echo $f
  sed "s;${pkgdir};\${SCN_FQ_DIR};" $f > ${f}.new
  rm $f
  mv ${f}.new $f
done

# Build SparseConvNet from source.

mkdir SparseConvNet
tar xf ${tardir}/${pkgtarname}.tar -C SparseConvNet || exit 1
cd SparseConvNet
python setup.py install

# Remove __pycache__'s.

find $pkgdir -name __pycache__ -exec rm -rf {} \; -prune

set +x

# real ups declare
cd ${product_dir}
${SSIBUILDSHIMS_DIR}/bin/declare_product ${product_dir} ${package} ${pkgver} NO_QUAL || \
  ssi_die "failed to declare ${package} ${pkgver} ${fullqual}"

# -------------------------------------------------------------------
# common bottom stuff
# -------------------------------------------------------------------

# this should not complain
echo "Finished building ${package} ${pkgver}"
cd ${product_dir}
setup ${package} ${pkgver} -z ${product_dir}:${PRODUCTS}

# this must be last
if [ "${maketar}" = "tar" ]
then
   echo ${SSIBUILDSHIMS_DIR}/bin/make_distribution_tarball ${product_dir} ${package} ${pkgver} NO_QUAL
   ${SSIBUILDSHIMS_DIR}/bin/make_distribution_tarball ${product_dir} ${package} ${pkgver} NO_QUAL
fi

exit 0
